#pragma once
#include <stdint.h>
#include <string>
#include "rxmesh/local.h"
#include "rxmesh/patch_info.h"
#include "rxmesh/util/macros.h"

namespace rxmesh {

namespace detail {
/**
 * @brief Return unique index of the local mesh element composed by the
 * patch id and the local index
 *
 * @param local_id the local within-patch mesh element id
 * @param patch_id the patch owning the mesh element
 * @return
 */
uint64_t __device__ __host__ __forceinline__ unique_id(const uint16_t local_id,
                                                       const uint32_t patch_id)
{
    uint64_t ret = patch_id;
    ret          = (ret << 32);
    ret |= local_id;
    return ret;
}

/**
 * @brief unpack a 64 uint to its high and low 32 bits. The low 32 bit are
 * casted to 16 bit. This is used to convert the unique id to its local id (16
 * low bit) and patch id (high 32 bit)
 * @param uid unique id
 * @return a std::pair storing the patch id and local id
 */
std::pair<uint32_t, uint16_t> __device__ __host__ __forceinline__
unpack(uint64_t uid)
{
    uint16_t local_id = uid & ((1 << 16) - 1);
    uint32_t patch_id = uid >> 32;
    return std::make_pair(patch_id, local_id);
}

}  // namespace detail

/**
 * @brief vertex identifier. It is a unique handle for each vertex equipped with
 * operator==. It can be used to access mesh (vertex) attributes
 */
struct VertexHandle
{
    using LocalT = LocalVertexT;

    /**
     * @brief Default constructor
     */
    __device__ __host__ VertexHandle() : m_handle(INVALID64)
    {
    }

    /**
     * @brief Constructor meant to be used internally by RXMesh and
     * query_dispatcher
     * @param patch_id the patch where the vertex belongs
     * @param vertex_local_id the vertex local index within the patch
     */
    __device__ __host__ VertexHandle(uint32_t     patch_id,
                                     LocalVertexT vertex_local_id)
        : m_handle(detail::unique_id(vertex_local_id.id, patch_id))
    {
    }

    /**
     * @brief Operator ==
     */
    bool __device__ __host__ __inline__ operator==(
        const VertexHandle& rhs) const
    {
        return m_handle == rhs.m_handle;
    }

    /**
     * @brief Operator !=
     */
    bool __device__ __host__ __inline__ operator!=(
        const VertexHandle& rhs) const
    {
        return !(*this == rhs);
    }

    /**
     * @brief Check if the vertex is valid i.e., has been initialized by RXMesh
     */
    bool __device__ __host__ __inline__ is_valid() const
    {
        return m_handle != INVALID64;
    }

    /**
     * @brief The unique identifier that represents the vertex
     */
    uint64_t __device__ __host__ __inline__ unique_id() const
    {
        return m_handle;
    }

    /**
     * @brief Unpack the handle to its patch id and vertex local index within
     * the patch
     */
    std::pair<uint32_t, uint16_t> __device__ __host__ __inline__ unpack() const
    {
        return detail::unpack(m_handle);
    }

   private:
    uint64_t m_handle;
};

/**
 * @brief print vertex unique_id to ostream
 */
inline std::ostream& operator<<(std::ostream& os, VertexHandle v_handle)
{
    return (os << 'v' << v_handle.unique_id());
}

/**
 * @brief edge identifier. It is a unique handle for each edge equipped with
 * operator==. It can be used to access mesh (edge) attributes
 */
struct EdgeHandle
{
    using LocalT = LocalEdgeT;

    /**
     * @brief Default constructor
     */
    __device__ __host__ EdgeHandle() : m_handle(INVALID64)
    {
    }

    /**
     * @brief Constructor meant to be used internally by RXMesh and
     * query_dispatcher
     * @param patch_id the patch where the edge belongs
     * @param edge_local_id the edge local index within the patch
     */
    __device__ __host__ EdgeHandle(uint32_t patch_id, LocalEdgeT edge_local_id)
        : m_handle(detail::unique_id(edge_local_id.id, patch_id))
    {
    }

    /**
     * @brief Operator ==
     */
    bool __device__ __host__ __inline__ operator==(const EdgeHandle& rhs) const
    {
        return m_handle == rhs.m_handle;
    }


    /**
     * @brief Operator !=
     */
    bool __device__ __host__ __inline__ operator!=(const EdgeHandle& rhs) const
    {
        return !(*this == rhs);
    }

    /**
     * @brief Check if the edge is valid i.e., has been initialized by RXMesh
     */
    bool __device__ __host__ __inline__ is_valid() const
    {
        return m_handle != INVALID64;
    }

    /**
     * @brief The unique identifier that represents the edge
     */
    uint64_t __device__ __host__ __inline__ unique_id() const
    {
        return m_handle;
    }

    /**
     * @brief Unpack the handle to its patch id and edge local index within
     * the patch
     */
    std::pair<uint32_t, uint16_t> __device__ __host__ __inline__ unpack() const
    {
        return detail::unpack(m_handle);
    }

   private:
    uint64_t m_handle;
};
/**
 * @brief print edge unique_id to ostream
 */
inline std::ostream& operator<<(std::ostream& os, EdgeHandle e_handle)
{
    return (os << 'e' << e_handle.unique_id());
}

/**
 * @brief face identifier. It is a unique handle for each face equipped with
 * operator==. It can be used to access mesh (face) attributes
 */
struct FaceHandle
{
    using LocalT = LocalFaceT;

    /**
     * @brief Default constructor
     */
    __device__ __host__ FaceHandle() : m_handle(INVALID64)
    {
    }

    /**
     * @brief Constructor meant to be used internally by RXMesh and
     * query_dispatcher
     * @param patch_id the patch where the face belongs
     * @param vertex_local_id the face local index within the patch
     */
    __device__ __host__ FaceHandle(uint32_t patch_id, LocalFaceT face_local_id)
        : m_handle(detail::unique_id(face_local_id.id, patch_id))
    {
    }

    /**
     * @brief Operator ==
     */
    bool __device__ __host__ __inline__ operator==(const FaceHandle& rhs) const
    {
        return m_handle == rhs.m_handle;
    }

    /**
     * @brief Operator !=
     */
    bool __device__ __host__ __inline__ operator!=(const FaceHandle& rhs) const
    {
        return !(*this == rhs);
    }

    /**
     * @brief Check if the face is valid i.e., has been initialized by RXMesh
     */
    bool __device__ __host__ __inline__ is_valid() const
    {
        return m_handle != INVALID64;
    }

    /**
     * @brief The unique identifier that represents the face
     */
    uint64_t __device__ __host__ __inline__ unique_id() const
    {
        return m_handle;
    }

    /**
     * @brief Unpack the handle to its patch id and face local index within
     * the patch
     */
    std::pair<uint32_t, uint16_t> __device__ __host__ __inline__ unpack() const
    {
        return detail::unpack(m_handle);
    }

   private:
    uint64_t m_handle;
};

/**
 * @brief print face unique_id to ostream
 */
inline std::ostream& operator<<(std::ostream& os, FaceHandle f_handle)
{
    return (os << 'f' << f_handle.unique_id());
}
}  // namespace rxmesh